(*
 *  beware IRC services, b_gline.pas
 *  Copyright (C) 2002 Bas Steendijk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)
unit b_gline;

interface

{
G-line code for operator service
}

uses buser,bcmds,bstuff,blinklist,bpremaskmatch,unitbanmask;

const
  glineflag_active=   $1;
  glineflag_local=    $2;
  glineflag_badchan=  $4;
  glineflag_ipmask=   $8;

type
  tgline=class(tlinklist)
    sender:string;
    mask,umask:string;
    m:tbanmask;
    date:integer;
    reason:string;
    flags:integer;
  end;

var
  glist:tlinklist;
  badchanglist:tlinklist;

procedure timehandler;
procedure cleargline;

procedure addgline(sender:string;mask:string;ts:integer;reason:string);

procedure mc_gline(cptr,sptr:tuser;parc:integer;parv:pparams);
function validglinereason(const s:string):boolean;
function validglinemask(var s:string):boolean;
procedure sendgline(gl:tgline);
function findgline(const s:string):tgline;
procedure delgline(sptr:tuser;mask:string);

procedure savegline(const filename:string);
procedure loadgline(const filename:string);

implementation

uses btime,bchannel,bparse,bconfig;

function validglinereason(const s:string):boolean;
begin
  {G-line reason must be 2 words}
  result := false;
  if length(s) < 3 then exit;
  if pos(' ',s) = 0 then exit;
  if s[1] <= #32 then exit;
  if s[length(s)] <= #32 then exit;
  result := true;
end;

function validglinemask(var s:string):boolean;
var
  a,b:integer;
begin
  {
  - can modify the mask (add *@)

  - atleast 4 in A-Z,a-z,0-9

  }
  result := false;
  if s = '' then exit;
  if pos('+',s) <> 0 then exit;
  if pos('-',s) <> 0 then exit;
  if pos('@',s) = 0 then begin
    if pos('.',s) = 0 then s := s + '@*' else s := '*@' + s;
  end;
  if s[1] = '@' then s := '*'+s;
  if s[length(s)] = '@' then s := s+'*';

  b := 0;
  for a := 1 to length(s) do begin
    if (s[a] in ['0'..'9']) or (s[a] in ['A'..'Z']) or (s[a] in ['a'..'z']) then inc(b);
  end;
  if b < 4 then exit;
  result := true;
end;

function findgline(const s:string):tgline;
var
  gl:tgline;
begin
  {return g-line with the same mask, s must already be uppercased}
  result := nil;
  if s = '' then exit;

  if ischanprefix(s[1]) then
  gl := tgline(badchanglist)
  else
  gl := tgline(glist);
  while gl <> nil do begin
    if gl.umask = s then begin
      result := gl;
      exit
    end;
    gl := tgline(gl.next);
  end;
end;

procedure sendgline(gl:tgline);
begin
  send(me.idstr+' '+TOK_GLINE+' * +'+gl.mask+' '+inttostr(gl.date-irctime)+' :'+gl.reason);
end;

procedure addgline(sender:string;mask:string;ts:integer;reason:string);
var
  gl,gl2,glfound:tgline;
  s:string;
  isbadchan:boolean;
  premask:tpremask;
  bm:tbanmask;
begin
  if ts <= 0 then exit;
  ts := irctime+ts;
  if ts < 0 then ts := maxlongint;

  if mask = '' then exit;

  if ischanprefix(mask[1]) then begin
    isbadchan := true;
  end else begin
    isbadchan := false;
    if pos('@',mask) = 0 then mask := '*@'+mask;
  end;

  {remove any G-line, overlapped but not the same}
  glfound := nil;
  s := ircupper(mask);
  gl := tgline(glist);
  banmaskmake(@bm,s);
  premaskmake(@premask,s);

  while gl <> nil do begin
    gl2 := tgline(gl.next);
    if banpremaskmatch(@premask,@bm,@gl.m,s,gl.umask) then begin
      if (s <> gl.umask) then begin
        if ts >= gl.date then begin
          linklistdel(glist,tlinklist(gl));
          gl.destroy;
        end;
      end else glfound := gl;
    end;
    gl := gl2;
  end;

  {function "prebanmaskmatch": combined premaskmatch and banmaskmatch, params: premask, banmask1, banmask2}
  (*
  if bm.bin then begin
    while gl <> nil do begin
      gl2 := tgline(gl.next);
      if gl.m.bin then begin
        if banmaskmatch(@bm,@gl.m) then if (s <> gl.umask) then if ts >= gl.date then begin
          linklistdel(glist,tlinklist(gl));
          gl.destroy;
        end;
      end else begin
        if premaskmatch(@premask,gl.umask) then
        if maskmatch(s,gl.umask) then if (s <> gl.umask) then if ts >= gl.date then begin
          linklistdel(glist,tlinklist(gl));
          gl.destroy;
        end;
      end;
      gl := gl2;
    end;
  end else begin
    while gl <> nil do begin
      gl2 := tgline(gl.next);
      {a non IP gline can't overlap IP G-lines, no need to check}
      if not gl.m.bin then begin
        if premaskmatch(@premask,gl.umask) then
        if maskmatch(s,gl.umask) then if (s <> gl.umask) then if ts >= gl.date then begin
          linklistdel(glist,tlinklist(gl));
          gl.destroy;
        end;
      end;
      gl := gl2;
    end;
  end;
  *)

{  gl := findgline(s);}
  gl := glfound;
  if gl = nil then begin
    {add g-line}

      gl := tgline.create;
      if isbadchan then begin
        setflag(gl.flags,glineflag_badchan);
        linklistadd(badchanglist,tlinklist(gl));
      end else begin
        linklistadd(glist,tlinklist(gl));
      end;
      gl.mask := mask;
      gl.umask := s;
      gl.date := ts;
      gl.reason := reason;
      gl.sender := sender;
      if isipmask(gl.mask) then setflag(gl.flags,glineflag_ipmask);

      if not isbadchan then
      banmaskmake(@gl.m,gl.mask);
  end else begin
    {reset expiration date}

    gl.date := ts;
    gl.sender := sender;
    if validglinereason(reason) then gl.reason := reason;
  end;
  setflag(gl.flags,glineflag_active);

  sendgline(gl);
end;

procedure delgline(sptr:tuser;mask:string);
var
  gl:tgline;
  isbadchan:boolean;
begin

  if ischanprefix(mask[1]) then begin
    isbadchan := true;
  end else begin
    isbadchan := false;
    validglinemask(mask);
  end;

  gl := findgline(ircupper(mask));
  if gl <> nil then begin
    if isbadchan then begin
      linklistdel(badchanglist,tlinklist(gl));
    end else begin
      linklistdel(glist,tlinklist(gl));
    end;
    send(me.idstr+' '+TOK_GLINE+' * -'+gl.mask);
    gl.destroy;
  end;
end;

procedure mc_gline(cptr,sptr:tuser;parc:integer;parv:pparams);
var
  gl:tgline;
  sendtalk:string;
  a:integer;

procedure sendinfo(gl:tgline);
begin
  inc(a);
  send(sendtalk+gl.mask+' <'+timestrshort(gl.date)+' '+gl.sender+'> '+gl.reason);
end;

begin
  sendtalk := sendtalkstr(cptr,sptr);
  {gline by client. if allowed to see them, show G-line list}

  a := 0;

  if (parv[1] = '') or (parc < 2) then parv[1] := '*';
        gl := tgline(glist);
        if parv[2] = '*' then begin
          while gl <> nil do begin
            sendinfo(gl);
            gl := tgline(gl.next);
          end;
          gl := tgline(badchanglist);
          while gl <> nil do begin
            sendinfo(gl);
            gl := tgline(gl.next);
          end;

        end else begin
          parv[1] := ircupper(parv[1]);
          while gl <> nil do begin
            if maskmatch(parv[1],gl.umask) then sendinfo(gl);
            gl := tgline(gl.next);
          end;
          gl := tgline(badchanglist);
          while gl <> nil do begin
            if maskmatch(parv[1],gl.umask) then sendinfo(gl);
            gl := tgline(gl.next);
          end;
        end;
        send(sendtalk+'End of G-line list, '+inttostr(a)+' listed.');
end;

{called by the ontimer event once/sec}
procedure timehandler;
var
  gl,gl2:tgline;
begin
  if irctime and 7 <> 0 then exit; {only once per 8 secs is often enough}

  gl := tgline(glist);
  while gl <> nil do begin
    gl2 := tgline(gl.next);
    if gl.date < irctime then begin
      linklistdel(glist,tlinklist(gl));
      gl.destroy;
    end;
    gl := gl2;
  end;

  gl := tgline(badchanglist);
  while gl <> nil do begin
    gl2 := tgline(gl.next);
    if gl.date < irctime then begin
      linklistdel(badchanglist,tlinklist(gl));
      gl.destroy;
    end;
    gl := gl2;
  end;
end;

procedure cleargline;
var
  gl,gl2:tgline;
begin
  gl := tgline(glist);
  while gl <> nil do begin
    gl2 := tgline(gl.next);
    linklistdel(glist,tlinklist(gl));
    gl.destroy;
    gl := gl2;
  end;

  gl := tgline(badchanglist);
  while gl <> nil do begin
    gl2 := tgline(gl.next);
    linklistdel(badchanglist,tlinklist(gl));
    gl.destroy;
    gl := gl2;
  end;
end;

procedure savegline(const filename:string);
var
  t:textfile;
  gl:tgline;

procedure writegline;
begin
  writeln(t,'[gline]');
  writeln(t,'mask=',gl.mask);
  writeln(t,'expire=',gl.date);
  writeln(t,'sender=',gl.sender);
  writeln(t,'reason=',gl.reason);
  writeln(t);

end;

begin
  assignfile(t,filename);
  filemode := 2;
  {$i-}rewrite(t);{$i+}
  if ioresult <> 0 then begin
    {saving G-lines failed}
    exit;
  end;
  gl := tgline(glist);
  while gl <> nil do begin
    writegline;
    gl := tgline(gl.next);
  end;
  gl := tgline(badchanglist);
  while gl <> nil do begin
    writegline;
    gl := tgline(gl.next);
  end;
  closefile(t);
end;

var
  currentgline:tgline;

procedure loadglinefinish;
begin
  if currentgline <> nil then begin
      if currentgline.mask <> '' then begin
        if ischanprefix(currentgline.mask[1]) then begin
          setflag(currentgline.flags,glineflag_badchan);
          linklistadd(badchanglist,tlinklist(currentgline));
        end else begin
          banmaskmake(@currentgline.m,currentgline.mask);
          clearflag(currentgline.flags,glineflag_badchan);
          linklistadd(glist,tlinklist(currentgline));
          {validglinemask(sn);}
        end;
      end else currentgline.destroy;
  end;
end;

procedure loadglineproc(sv,sn:string);
begin
  if sv = '[GLINE]' then begin
    loadglinefinish;
    currentgline := tgline.create;
    setflag(currentgline.flags,glineflag_active);
  end;
  if currentgline = nil then exit;
  if sv = 'MASK' then begin
    currentgline.mask := sn;
    currentgline.umask := ircupper(sn);
  end;
  if sv = 'EXPIRE' then currentgline.date := strtointdef(sn,0);
  if sv = 'SENDER' then currentgline.sender := sn;
  if sv = 'REASON' then currentgline.reason := sn;
end;

procedure loadgline(const filename:string);
begin
  cleargline;
  currentgline := nil;
  readcfg(filename,@loadglineproc);
  loadglinefinish;
  currentgline := nil;

  {remove any expired g-lines}
  timehandler;
end;

begin
  glist := nil;
  badchanglist := nil;
end.
