(*
 *  beware IRC services, boptiontable.pas
 *  Copyright (C) 2002 Bas Steendijk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)
unit boptiontable;

interface

{
string name: subname?

idea: give load and save a [name]

load: searches for [name], parses options until next [* line or EOF

save: searches for [name], copying everything it finds, writes, copies from next [* line


------------------
save calls save custom
save custom: count, callback. save custom loops from 0 to max,
calls callback string function with number param
}

const
  opt_int=1;
  opt_bool=2;
  opt_str=3;

type
  toptionentry=record
    p:pointer;   {pointer to option var}
    typ:integer;   {type of option var}
    name:string; {id name of option}
    info:string;  {description}
    min,max:integer;
    def_s:string; {default if string}
    case integer of
      0: (def_i:integer); {default if integer}
      1: (def_b:boolean) {default if boolean}
  end;
  toptiontable=array[0..0] of toptionentry;
  poptiontable=^Toptiontable;

procedure optiontable_default(var v;count:integer);
function optiontable_load(var v;count:integer;filename,subname:string):boolean;
procedure optiontable_save(var v;count:integer;filename,subname:string);
function parseoptions(var v;count:integer;s:string):boolean;
function optionstr(var v;num:integer):string;

implementation

uses readtxt,bstuff;

function parseoptions(var v;count:integer;s:string):boolean;
var
  s2,sv,sn:string;
  a,b:integer;
  optiontable:toptiontable absolute v;
begin
  result := false;
  if s = '' then exit;
  if s[1] = ';' then exit;
  if s[1] = ' ' then exit;
  while s[length(s)] = ' ' do setlength(s,length(s)-1); {strip off spaces at the end}

  s2 := ircupper(s);
  a := pos('=',s);
  if a <> 0 then begin
    sv := copy(s2,1,a-1);
    sn := copy(s,a+1,length(s))
  end;

  for a := 0 to count do if ircupper(optiontable[a].name) = sv then begin
    case optiontable[a].typ of
      opt_int:begin
        b := strtointdef(sn,integer(optiontable[a].p^));
        if (optiontable[a].min <> 0) and (optiontable[a].max <> 0) then begin
          if b < optiontable[a].min then b := optiontable[a].min;
          if b > optiontable[a].max then b := optiontable[a].max;
        end;
        result := b <> integer(optiontable[a].p^);
        integer(optiontable[a].p^) := b;
      end;
      opt_bool:begin
        if sn = '0' then begin
          result := boolean(optiontable[a].p^);
          boolean(optiontable[a].p^) := false
        end else if sn = '1' then begin
          result := not boolean(optiontable[a].p^);
          boolean(optiontable[a].p^) := true;
        end;
      end;
      opt_str:begin
        result := sn <> string(optiontable[a].p^);
        string(optiontable[a].p^) := sn;

      end;
    end;
    break;
  end;

end;

procedure optiontable_default(var v;count:integer);
var
  a:integer;
  optiontable:toptiontable absolute v;
begin
  for a := 0 to count do begin
    case optiontable[a].typ of
      opt_int:integer(optiontable[a].p^) := optiontable[a].def_i;
      opt_bool:boolean(optiontable[a].p^) := optiontable[a].def_b;
      opt_str:string(optiontable[a].p^) := optiontable[a].def_s;
    end;
  end;
end;

function optiontable_load(var v;count:integer;filename,subname:string):boolean;
var
  pt:treadtext;
  optiontable:toptiontable absolute v;
begin
  result := false;
  if not readtext_init(pt,filename) then exit;
  result := true;
  while not readtext_eof(pt) do parseoptions(optiontable,count,readtext_line(pt));
  readtext_close(pt);
end;

procedure optiontable_save(var v;count:integer;filename,subname:string);
var
  t:textfile;
  a:integer;
  optiontable:toptiontable absolute v;
begin
  assignfile(t,filename);
  filemode := 2;
  {$i-}rewrite(t);{$i+}
  if ioresult <> 0 then exit;

  if subname <> '' then writeln(t,subname);
  {subname can be used for more advanced stuff than this}

  for a := 0 to count do begin
    writeln(t,optiontable[a].name,'=',optionstr(optiontable,a));
    if optiontable[a].info <> '' then begin
      writeln(t,';',optiontable[a].info);
      writeln(t);
    end;
  end;
  closefile(t);
end;

function optionstr(var v;num:integer):string;
var
  optiontable:toptiontable absolute v;
begin
  case optiontable[num].typ of
    opt_int:result := inttostr(integer(optiontable[num].p^));
    opt_bool:result := inttostr(ord(boolean(optiontable[num].p^)));
    opt_str:result := string(optiontable[num].p^);
  end;
end;

end.
