(*
 *  beware IRC services, bparse.pas
 *  Copyright (C) 2002 Bas Steendijk
 *
 *  some modifications by peter green
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

 {most P10 stuff is to be in this unit for easy porting to another server protocol}

unit bparse;

interface

uses bservicemngr,buser,pgdebugout;

var
  isconnected:boolean;
  online:boolean;
  mynum:string;
  hostservername:string;
  sendqsize:integer;

procedure send(const s:string);
procedure sendflush;
procedure sendburst;
procedure onconnected;
procedure parsecommand(mine:boolean;const s:string);
procedure sendbot_nick(us:tuser);
procedure sendparse(const s:string);
procedure sendkill(target:tuser;const fromstr,reason:string);

{
i wrap everything so a bot can without much effort be used on a different server-server protocol
}
function sendprefixstr(from:tuser):string;
function sendtalkstr(from,target:tuser):string;
procedure bot_join(sptr:tuser;channame:string;ops:boolean;ts:integer);
procedure bot_quit(sptr:tuser;reason:string);

procedure onnewuser(us:tuser);

implementation

uses
  {winsock,wsocket,}
  bservunit,bstuff,sysutils,btime,bconsts,bcmds,bircrandom,
  bconfig,bchannel,bserver,bcrash,b_burst,b_clearmode,bipcheck;

function sendtalkstr(from,target:tuser):string;
begin
  result := sendprefixstr(from)+TOK_NOTICE+' '+target.idstr+' :';
end;

{gives the ":sender " part of a message to send}
function sendprefixstr(from:tuser):string;
begin
  result := from.idstr+' ';
end;

procedure sendbot_nick(us:tuser);
begin
  send(mynum+' '+TOK_NICK+' '+us.name+' 1 '+inttostr(globaldefaultts)+' '+us.userid+' '+us.host+usermodestr(us)+' '+p10inttostr(us.binip,6)+' '+us.idstr+' :'+us.fullname);
end;

procedure sendkill(target:tuser;const fromstr,reason:string);
begin
  sendparse(mynum+' '+TOK_KILL+' '+target.idstr+' :'+me.name+'!'+fromstr+' ('+reason+')');
end;

procedure sendflush;
begin
  if not isconnected then exit;
  {$ifndef win32}
    sock.sendflush;
  {$else}
    sock.send(nil,0);
  {$endif}
  sendqsize := 0; 
  
end;

procedure send(const s:string);
begin
  if isconnected then begin
    inc(sendqsize,length(s)+2);
    sock.putstringinsendbuffer(s+#13#10);
    {sock.sendline(s);}
    if sendqsize > 1500 then sendflush;
  end;
end;

procedure bot_join(sptr:tuser;channame:string;ops:boolean;ts:integer);
begin
  sendparse(sptr.idstr+' '+TOK_JOIN+' '+channame+' '+inttostr(irctime));
  if ops then sendparse(me.idstr+' '+TOK_MODE+' '+channame+' +o '+sptr.idstr+' '+inttostr(ts));
end;

procedure bot_quit(sptr:tuser;reason:string);
begin
  sendparse(sptr.idstr+' '+TOK_QUIT+' :'+reason);
end;


procedure onconnected;
begin
  send('PASS '+opt.connectpass);
  send('SERVER '+me.name+' 1 '+inttostr(irctime)+' '+inttostr(irctime)+' J10 '+mynum+p10inttostr(maxbot,CCClen)+' :'+opt.info);
end;

procedure sendburst;
var
  a,c:integer;
  ch:tchannel;
  uc:tuserchan;
  ban:tban;
  bool,goed:boolean;
  s:string;
begin
  {send all bots already in list}
  for a := 0 to maxbot do if botarray[a] <> nil then begin
    sendbot_nick(botarray[a]);
  end;

  {sendburst all currently existing channels}
  ch := tchannel(globalchanlist);
  while ch <> nil do begin

      s := '';
      uc := tuserchan(ch.user);
      ban := ch.banlist;
      goed := true;

      repeat
        if goed then begin
          s := chanmodestr(ch,true);
          goed := false;
        end else s := '';
        c := 0;
        bool := true;
        while (uc <> nil) and (length(s) < 400) do begin
          if bool then begin
            s := s + ' ';
            bool := false;
          end else s := s + ',';
          s := s + uc.us.idstr;
          if uc.flags and (userchanflag_op or userchanflag_voice) <> c then begin
            c := uc.flags and (userchanflag_op or userchanflag_voice);
            case c of
              0:s := s + ':';
              userchanflag_voice:s := s + ':v';
              userchanflag_op:s := s + ':o';
              userchanflag_op or userchanflag_voice:s := s + ':ov';
            end;
          end;
          uc := tuserchan(uc.next2);
        end;
        bool := true;
        while (ban <> nil) and (length(s) < 400) do begin
          if bool then begin
            s := s + ' :%';
            bool := false;
          end;
          s := s + ban.mask+' ';
          ban := tban(ban.next);
        end;
        send(me.idstr+' '+TOK_BURST+' '+ch.name+' '+inttostr(ch.ts)+s);
      until (uc = nil) and (ban = nil);

    ch := tchannel(ch.next);
  end;

  {call onsendburst of all modules}

  send(mynum+' '+TOK_END_OF_BURST);
end;


{
send something, and parse it - the bots see the message being sent
(a NOTICE can be sent w/o parsing it)

cpu consuming but easy way to keep everything synched

example:

chanserv -> MODE change
scanner -> PRIVMSG operserv GLINE
}
procedure sendparse(const s:string);
begin
  send(s);
  parsecommand(true,s);
end;

procedure onnewuser(us:tuser);
var
  a:integer;
  s:tserviceunit;
begin
  ipcheck_remoteclient(us);
  for a := maxserviceunit downto 0 do begin
    s := serviceunitarray[a];
    if s <> nil then if @s.onnewuser <> nil then s.onnewuser(us,s.tag);
  end;
end;

procedure onaccount(sptr:tuser;us:tuser);
var
  a:integer;
  s:tserviceunit;
begin

  for a := maxserviceunit downto 0 do begin
    s := serviceunitarray[a];
    if s <> nil then if @s.onaccount <> nil then s.onaccount(sptr,us,s.tag);
  end;

end;


procedure parsecommand(mine:boolean;const s:string);
var
  a,b,c,d:integer;
  cmd,cmd2:string;
  parc:integer;
  newcommand:boolean;
  parv,parv2:tparams;
  bool:boolean;
  sender:string;
  s1:string;
  sptr:tuser;
  us:tuser;
  ch:tchannel;
  uc:tuserchan;
begin
 try

  debugout('parsecommand entered');

  if s = '' then exit;
  if length(s) > maxmessagelength then exit;

  {fs.memo1.lines.add(s);}

  if s[1] = ':' then begin
    newcommand := false;
    a := pos(' ',s);
    sender := copy(s,2,a-2);
  end else begin
    if (online or mine) then begin
      newcommand := true;
      a := pos(' ',s);
      sender := copy(s,1,a-1);
    end else begin
      newcommand := false;
      sender := '';
      a := 1;
    end;
  end;
  if a = 0 then exit;
  if sender <> '' then sender := convertidstr(sender);
  if sender <> '' then if (a >= length(s)) then exit;

  {a is index in parsed string - get index of start of command}
  while s[a] = ' ' do inc(a);
  b := 0;
  c := length(s);
  bool := false;
  while (a <= c) do begin
    if b > mparams then break;
    if s[a] = ':' then begin
      bool := true;
      break;
    end;
    d := a;
    while (s[a] <> ' ') and (a <= c) do inc(a);
    parv[b] := copy(s,d,a-d);
    inc(b);
    while (s[a] = ' ') and (a <= c) do inc(a)
  end;
  inc(a);
  if bool then begin
    parv[b] := copy(s,a,c);
    inc(b)
  end;

  parc := b;
  if parc = 0 then exit; {no command, only a :sender prefix}

  cmd2 := parv[0];
  cmd := ircupper(cmd2);

  if not mine then if cmd = MSG_ERROR then begin
    {fs.caption := parv[parc-1];}
    if pos('throttled',parv[parc-1]) <> 0 then begin
      reconnectcountdown := 45;
    end;
  end;

  {unknown command}

  if not mine then if copy(sender,1,SSlen) = me.idstr then exit;
  if not (mine or online) then begin
    if cmd = 'SERVER' then begin
      us := adduser;
      hostservuser := us;
      setnickname(us,parv[1]);
      us.hops := 1;
      us.fullname := copy(parv[8],1,maxgcoslength);
      parv[6] := convertidstr(parv[6]);

      assert((length(parv[6]) = SSCCClen) and (p10strtoint(parv[6]) >= 0) ,'parse: SERVER: invalid numeric',s);

      us.idstr := copy(parv[6],1,SSlen);

      assert(p10server[p10strtoint(us.idstr)] = nil,'parse: SERVER: server numeric collision',s);
      addserver(us,p10strtoint(us.idstr),p10strtoint(copy(parv[6],1+SSlen,CCClen)),me.server);

      online := true;
      sendburst;
    end;
    exit;
  end;

  if newcommand then begin
    {prefix had no colon, its a numeric}
    sptr := findnumeric(sender);
  end else begin
    {prefix had a colon, it's a name}
    sptr := findname(sender);
  end;

  if sptr = nil then exit;
  if cmd = TOK_PING then begin
    send(mynum+' '+TOK_PONG+' '+me.idstr+' :'+parv[1]);
  end else if cmd = TOK_ACCOUNT then begin
    us := findnumeric(parv[1]);

    {assert(assigned(us),'parse: ACCOUNT: user not found',s);}
    if assigned(us) then begin
      if parv[2] <> '' then begin
        us.account := parv[2];
        onaccount(sptr,us);
      end;
    end;
  end else if cmd = TOK_NICK then begin
    if isserver(sptr) then begin
      {server introducing new client}

      parv[parc-2] := convertidstr(parv[parc-2]);
      assert((length(parv[parc-2]) = SSCCClen) and (p10strtoint(parv[parc-2]) >= 0) ,'parse: NICK: invalid numeric',s);

      a := p10strtoint(copy(parv[parc-2],1,SSlen));
      b := p10strtoint(copy(parv[parc-2],1+SSlen,CCClen));
      assert((a >= 0) and (b >= 0),'parse: N: bogus numeric',s);
      assert(p10server[a] <> nil,'parse: N: servernum does not exist',s);
      b := b and p10server[a].p10max;

      assert(p10server[a].p10slots[b] = nil,'parse: N: numeric collision',s);

      {nick collision?}
      us := findname(parv[1]);
      if us <> nil then begin
        assert(us.server = me.server,'parse: N: nick collision, both users are not a service',s);
        exit;
      end;

      us := adduser;
      p10server[a].p10slots[b] := us;
      us.p10slotnum := b;
      setnickname(us,parv[1]);
      us.ts := strtointdef(parv[3],0);
      us.hops := strtointdef(parv[2],0);
{      assert(us.hops = sptr.hops,'parse: N: client hopcount isn''t same as server hopcount',s);}
      us.userid := parv[4];
      us.host := parv[5];
      us.uhost := ircupper(parv[6]);
      a := p10strtoint(parv[parc-3]);
      us.idstr := parv[parc-2];
      us.binip := a;
      us.server := sptr.server;
      setflag(us.flags,userflag_isclient);
      inc(count.globalclients);

      if parv[6,1] = '+' then begin
      {modes}
        for a := 0 to maxusermodetable do if pos(usermodetable[a].c,parv[6]) <> 0 then setflag(us.modeflag,usermodetable[a].flag);
      end;

      if isoper(us) then inc(count.oper);
      if flag_isset(us.modeflag,usermode_invisible) then inc(count.invisible);
      us.fullname := parv[parc-1];

      {also called if it is mine?}
      onnewuser(us);
      if parv[6,1] = '+' then begin
        {search for mode +r (account burst) stores the login name
        and then passes the message on to any interested bots}
        if pos('r',parv[6]) <>0 then begin
          us.account := parv[7];
          onaccount(sptr,us);
        end;
      end;
    end else begin
      {client changing nick}
      setnickname(sptr,parv[1]);
      sptr.ts := strtointdef(parv[2],0);
    end;
  end else
  if cmd = TOK_SVSNICK then begin
    (*
    -- must not parse svsnick, new svsnick code
    us := findnumeric(parv[1]);
    if us = nil then exit;
    us2 := findname(parv[2]);

    assert((us2 = nil) or (us2 = us),'parse: SVSNICK: nick collision',s);
    setnickname(us,parv[2]);
    us.ts := strtointdef(parv[3],0);
    *)
  end else
  if cmd = TOK_QUIT then begin
    ircrandom; {call ircrandom to feed randomness}
    sptr.destroy;
  end else
  if cmd = TOK_KILL then begin
    us := findnumeric(parv[1]);

    {check - one of my bots?}
{    assert(us <> nil,'parse: KILL: no such user',s);}
    if assigned(us) then us.destroy;
  end else
  if cmd = TOK_SQUIT then begin
    us := findname(parv[1]);
    if us = me then us := hostservuser;
    assert(us <> nil,'parse: SQUIT: no such server',s);
    us.destroy;
  end else
  if cmd = TOK_SERVER then begin
    {introducing new server}
    parv[6] := convertidstr(parv[6]);
    assert((length(parv[6]) = SSCCClen) and (p10strtoint(parv[6]) >= 0) ,'parse: SERVER: invalid numeric',s);

    us := adduser;
    setnickname(us,parv[1]);
    us.hops := strtointdef(parv[2],0);
    us.fullname := copy(parv[parc-1],1,maxgcoslength);
    us.idstr := copy(parv[6],1,SSlen);
    assert(p10server[p10strtoint(us.idstr)] = nil,'parse: S: server numeric collision',s);
    addserver(us,p10strtoint(us.idstr),p10strtoint(copy(parv[6],1+SSlen,CCClen)),sptr.server);
    assert(us.hops = sptr.hops+1,'parse: S: bogus hopcount: '+inttostr(sptr.hops)+' '+inttostr(us.hops),s);
  end else
  if cmd = TOK_CREATE then begin
    assert(isclient(sptr),'parse: C',s);
    b := strtok(parv[1],',',@parv2);
    for a := 0 to b-1 do begin
      ch := findchan(parv2[a]);
      if not isonchannel(sptr,ch) then begin
        if ch = nil then begin
          ch := createchannel;
          setchanname(ch,parv2[a]);
          if ch.name[1] = '+' then setflag(ch.flags,chanflag_modeless);
        end;
        c := strtointdef(parv[2],0);
        if ((c < ch.ts) and (c >= OLDEST_TS)) or
        ((ch.ts < OLDEST_TS) and (c >= OLDEST_TS)) then ch.ts := c;
        uc := addusertochannel(sptr,ch);
      end else uc := getuserchan(sptr,ch);
      if not flag_isset(ch.flags,chanflag_modeless) then begin
        setflag(uc.flags,userchanflag_op);
      end;
    end;
  end else
  if cmd = TOK_JOIN then begin
    ircrandom; {call ircrandom to feed randomness}
    assert(isclient(sptr),'parse: J',s);
    b := strtok(parv[1],',',@parv2);
    for a := 0 to b-1 do begin
      ch := findchan(parv2[a]);
      if not isonchannel(sptr,ch) then begin
        if ch = nil then begin
          ch := createchannel;
          setchanname(ch,parv2[a]);
          if ch.name[1] = '+' then setflag(ch.flags,chanflag_modeless);
        end;
        c := strtointdef(parv[2],0);
        if ((c < ch.ts) and (c >= OLDEST_TS)) or
        ((ch.ts < OLDEST_TS) and (c >= OLDEST_TS)) then ch.ts := c;
        {uc := }addusertochannel(sptr,ch);
      end;
    end;
  end else
  if cmd = TOK_PART then begin
    b := strtok(parv[1],',',@parv2);
    for a := 0 to b-1 do begin
      ch := findchan(parv2[a]);
      if ch <> nil then begin
        uc := getuserchan(sptr,ch);
        if uc <> nil then deluserfromchannel(sptr,ch,uc);
      end;
    end;
  end else
  if cmd = TOK_KICK then begin
    us := findnumeric(parv[2]);
    ch := findchan(parv[1]);
    uc := getuserchan(us,ch);
    if uc <> nil then deluserfromchannel(us,ch,uc);
  end else
  if (cmd = TOK_MODE) or (cmd = TOK_OPMODE) then begin
    ch := findchan(parv[1]);
    if ch = nil then begin
      us := findnick(parv[1]);
      if us <> nil then begin
        setusermode(us,hostservuser,sptr,parc,@parv);
      end;
      exit;
    end;
    setchanmode(ch,hostservuser,sptr,parc,@parv);
  end else if (cmd = TOK_CLEARMODE) then begin
    m_clearmode(sptr,sptr,parc,@parv);
  end else
  if cmd = TOK_PRIVMSG then begin
    a := pos('@',parv[1]);
    if a <> 0 then begin
      s1 := copy(parv[1],a+1,500);
      if not strcompup(s1,me.name) then exit; {not at my server}
      s1 := copy(parv[1],1,a-1);
      us := findname(s1);
      if us = nil then exit; {no such nick}
      bool := true;
    end else begin
      bool := false;
      us := findnumeric(parv[1]);
      if us = nil then exit;
    end;
    if us.srvunit = nil then exit; {this user isnt my service bot, ignore message}
    tserviceunit(us.srvunit).onprivmsg(bool,sptr,us,parv[parc-1], tserviceunit(us.srvunit).tag);
  end else
  if cmd = TOK_BURST then begin
    m_burst(hostservuser,sptr,parc,@parv);

    {parse burst}
  end else
  if cmd = TOK_END_OF_BURST then begin
    if sptr = hostservuser then send(mynum+' '+TOK_EOB_ACK);
  end else if cmd=TOK_ACCOUNT then begin
  end;


  debugout('parsecommand complete');

 except

   assert(false,'exception raised in bparsecommand while parseing command '+s,exceptobject.classname+' '+(exceptobject as exception).message);
   
 end;
end;

begin

end.
