(*
 *  beware IRC services, bsignal.pas
 *  Copyright (C) 2002 Bas Steendijk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

unit bsignal;

{receive signals like restart, die, rehash}

interface

uses
  pgdebugout,
  {$ifdef win32}
    winsock,
    wsocket;
  {$else}
    lsocket,
    linux,
    sysutils,
    dos;   //despite the name the dos unit is available on almost all ports of fpc
  {$endif}

type
  tsignalclass=class
    {this procedure must be in a class}
    procedure ondataavailable(sender:tobject;error:word);


  end;


procedure signalstart;
procedure signalstop;
procedure sendsignal(s:string);
procedure sendsignal2;

var
  {$ifdef win32}
    signalsource:twsocket;
  {$else}
    signalsource:tlasin;
  {$endif}
implementation

uses bstuff,bconfig,{bwinnt,}bparse,buser,bcmds,bservicemngr,bconsts,bservunit;


var
  sc:tsignalclass;
  {$ifndef win32}
    recvq:string;
  {$endif}

procedure processsignal(s: string);
var
   a:integer;

   mask:array[0..maxserviceunit] of boolean;
   globalaction:boolean;
   parc:integer;
   parv:tparams;
   len:integer;
begin
  debugout('processing signal :'+s);
  len := length(s);
  if Len >= 0 then begin
    s := ircupper(s);

    parc := strtok(s,' ',@parv);
    // note parv is already uppercased
    globalaction := false;
    fillchar(mask,sizeof(mask),0);
    if (parc >= 2) and (parv[1] <> '') then begin
      for a := 0 to maxserviceunit do if serviceunitarray[a] <> nil then
      if ircupper(serviceunitarray[a].name) = parv[1] then mask[a] := true;
    end else begin
      globalaction := true;
      for a := 0 to maxserviceunit do mask[a] := true;
    end;

    if (parv[0] = 'STOP') then begin
      for a := 0 to maxserviceunit do if serviceunitarray[a] <> nil then
      if mask[a] then
      if @serviceunitarray[a].onshutdown <> nil then
      serviceunitarray[a].onshutdown(serviceunitarray[a].tag);
    end else
    if (parv[0] = 'DIE') then begin
      globalshutdown;
      halt;
    end else
    if (parv[0] = 'REHASH') then begin
      if globalaction then bconfig.init;

      for a := 0 to maxserviceunit do if serviceunitarray[a] <> nil then
      if mask[a] then
      if @serviceunitarray[a].onrehash <> nil then
      serviceunitarray[a].onrehash(serviceunitarray[a].tag);

    end else
    if (parv[0] = 'SAVE') then begin
      if globalaction then bconfig.writecfg;

      for a := 0 to maxserviceunit do if serviceunitarray[a] <> nil then
      if mask[a] then
      if @serviceunitarray[a].onsave <> nil then
      serviceunitarray[a].onsave(serviceunitarray[a].tag);

    end else
    if (parv[0] = 'START') then begin
      for a := 0 to maxserviceunit do if serviceunitarray[a] <> nil then
      if mask[a] then
      if @serviceunitarray[a].onstartup <> nil then
      serviceunitarray[a].onstartup(serviceunitarray[a].tag);
    end;
  end;
end;
procedure tsignalclass.ondataavailable(Sender: TObject; Error: Word);
var
  {$ifdef win32}
    Buffer : array [0..1023] of char;
    Len    : Integer;
    Src    : TSockAddrIn;
    SrcLen : Integer;
  {$else}
    a,b,c,d:integer;

    goed:boolean;
  {$endif}
  s:string;

begin
  {$ifdef win32}
    SrcLen := SizeOf(Src);
    Len    := twsocket(sender).ReceiveFrom(@Buffer, SizeOf(Buffer), Src, SrcLen);
    setlength(s,len);
    move(buffer[0],s[1],len);
    processsignal(s);
  {$else}
    debugout('data recieved on pipe');

    recvq := recvq + signalsource.receivestr;

    goed := false;
    a := 1;
    b := length(recvq);
    d := 0;
    repeat
      if length(recvq)=0 then exit; // a squit has resulted in the connection
                                    // being broken and reset the recvq
      c := a;

      while (a <= b) and (recvq[a] <> #13) and (recvq[a] <> #10) do inc(a);

      if a <= b then begin
        s := copy(recvq,c,a-c);
        d := a;
        if s <> '' then processsignal(s);

      end else goed := true;

      inc(a);
      //writeln('goed');
    until goed;
    if d > 0 then delete(recvq,1,d);
    sendflush;
  {$endif}
end;

procedure signalstart;
begin
  if signalsource <> nil then exit;
  {$ifdef win32}
    if opt.signalport <> '0' then begin
      signalsource := twsocket.create(nil);
      signalsource.Proto := 'udp';
      signalsource.addr := '127.0.0.1';
      signalsource.port := opt.signalport;
      try
        signalsource.listen;
        signalsource.ondataavailable := sc.ondataavailable;
      except
        signalstop;
      end;
    end;
  {$else}
    if opt.signalpipe <> '' then begin

      signalsource := tlasin.create;
      if fileexists(opt.signalpipe) then deletefile(opt.signalpipe);
      //i use the mknod command as fpc doesn't appear to support the mknod function
      debugout('about to run mknod');

      exec('/bin/mknod',opt.signalpipe+' p');
      debugout('mknod complete');
      //writeln(opt.signalpipe);
      signalsource.s := fdopen(opt.signalpipe,Open_Rdwr or OPEN_NONBLOCK);
      debugout('fdopen complete');
      if signalsource.s >= 0 then begin

        //fcntl(signalsource.s,F_SETFL,);
        //debugout('fcntl complete');
        signalsource.ondataavailable := sc.ondataavailable;
        signalsource.connected := true;
      end else begin
        writeln('failed to create signal pipe');
        signalstop;
      end;
      debugout('signalstart complete');
    end;
  {$endif}
end;

procedure signalstop;
begin
  if signalsource <> nil then signalsource.destroy;
  signalsource := nil;
  {$ifndef win32}
    if fileexists(opt.signalpipe) then deletefile(opt.signalpipe);
  {$endif}
end;

procedure sendsignal(s:string);

{$ifdef win32}
  var
    sock:twsocket;
  begin
    sock := twsocket.create(nil);
    sock.proto := 'udp';
    sock.addr := '127.0.0.1';
    sock.port := opt.signalport;
    sock.connect;
    sock.sendstr(s);
    sock.destroy;
  end;
{$else}
  var
    t : text;
  begin
    writeln(opt.signalpipe);
    assignfile(t,opt.signalpipe);
    append(t);
    writeln(t,s);
    closefile(t);
  end;
{$endif}
procedure sendsignal2;
var
  s:string;
  a:integer;
begin
  s := '';
  for a := 2 to paramcount do if paramstr(a) <> '' then begin
    if s <> '' then s := s + ' ';
    s := s + paramstr(a);
  end;
  sendsignal(s);
end;

begin
  signalsource := nil;
end.
